// Project : K-means
// Author : Dario GENGA

#include "kmeans.h"


/// Create empty clusters for the universe
/// \param universe The universe who contains the cluster to create.
void initialize_clusters_array(kmeans* universe) {
    universe->clusters_array = malloc(sizeof(cluster) * universe->k);

    for (int i = 0; i < universe->k; i++) {
        universe->clusters_array[i] = cluster_create(NULL);
    }
}

/// Initialize the array of points in the universe.
/// \param universe The universe who contains the points.
void initialize_points_array(kmeans* universe) {
    universe->points_array = malloc(sizeof(point) * universe->nb_points);
    for (int i = 0; i < universe->nb_points; i++) {
        universe->points_array[i] = NULL;
    }
}

/// Create a point from the values in a string.
/// \param line The string who contains the data of the point.
/// \param dimensions The number of dimension in the point.
/// \return The point created from the string.
point* create_point_from_string(char *line, int dimensions) {
    // Skip the parsing if the line doesn't contain data
    if (strcmp(line, "\n") == 0) {
        return NULL;
    }

    int i = 0;
    point* p;
    char *token;
    const char separator[2] = SEPARATOR;
    float* data = malloc(sizeof(float) * dimensions);

    // Parse the line
    token = strtok(line, separator);
    while(token != NULL) {
        // Convert the string value to float, then save it
        float value = atof(token);
        data[i] = value;

        // Get the next value
        token = strtok(NULL, separator);
        i++;
    }
    // Verify the dimensions of the point
    if (i != dimensions) {
        printf("Bad dimensions for the point.\n");
        exit(EXIT_FAILURE);
    }

    // Create the point and return it
    p = point_create(data, dimensions);
    return p;
}

kmeans* kmeans_create_empty() {
    kmeans* universe = malloc(sizeof(kmeans));
    universe->points_array = NULL;
    universe->nb_points = 0;
    universe->k = 0;
    universe->dimensions = 0;
    universe->clusters_array = NULL;

    return universe;
}

kmeans* kmeans_create(int k, int dimensions, point** data, int nb_points) {
    kmeans* universe = malloc(sizeof(kmeans));
    universe->points_array = data;
    universe->nb_points = nb_points;
    universe->k = k;
    universe->dimensions = dimensions;
    universe->clusters_array = NULL;
    initialize_clusters_array(universe);

    return universe;
}

point* point_create(float* value, int dimensions) {
    point* p = malloc(sizeof(point));
    p->value = value;
    p->cluster = NULL;
    p->label = NULL;
    p->dimensions = dimensions;
    return p;
}

cluster* cluster_create(point* centroid) {
    cluster* c = malloc(sizeof(cluster));
    c->centroid = centroid;
    return c;
}

point** read_data(char* source_file, int dimensions, int* data_length) {
    char *line = NULL;
    size_t buffer_size = 0;
    ssize_t line_length;
    int point_index = 0;

    // Count the number of data (points)
    int nb_data = count_file_lines(source_file);
    point **data = malloc(sizeof(point) * nb_data);
    *data_length = nb_data;

    // Open the file and read it, line by line
    FILE *file = open_file(source_file, "r");
    while((line_length = getline(&line, &buffer_size, file)) != -1) {
        point* p = create_point_from_string(line, dimensions);

        if (p != NULL) {
            data[point_index] = p;
            point_index++;
        }
    }

    // Close the file and free the memory
    fclose(file);
    free(line);
    return data;
}

void read_custom_data_source(kmeans* universe, char* source_file) {
    char *line = NULL;
    size_t buffer_size = 0;
    ssize_t line_length;
    int current_line = 0;
    int dimensions = 0;
    int point_index = 0;

    // Count the number of data (points)
    int nb_data = count_file_lines(source_file) - 2; // The two first line are not for the data
    universe->nb_points = nb_data;
    initialize_points_array(universe);

    // Open the file and read it, line by line
    FILE *file = open_file(source_file, "r");
    while((line_length = getline(&line, &buffer_size, file)) != -1) {
        // Remove newline char at end of line
        remove_string_trailing_newline(line);

        // Get the number of dimensions for each data
        if (current_line == LINE_INDEX_DIMENSIONS) {
            dimensions = (int)strtol(line, NULL, 10);
            universe->dimensions = dimensions;
        }
        // Get the number of cluster
        if (current_line == LINE_INDEX_CLUSTER) {
            universe->k = (int)strtol(line, NULL, 10);
            initialize_clusters_array(universe);
        }
        // Retrieve the data
        if (current_line >= LINE_INDEX_CONTENT) {
            point* p = create_point_from_string(line, dimensions);

            if (p != NULL) {
                universe->points_array[point_index] = p;
                point_index++;
            }
        }

        current_line++;
    }

    // Close the file and free the memory
    fclose(file);
    free(line);
}

void write_data_output(kmeans *universe, char* output_file) {
    // Open the file
    FILE *file = open_file(output_file, "w");

    // Write the number of dimensions
    fprintf(file, "%d\n", universe->dimensions);
    // Write the number of clusters
    fprintf(file, "%d\n", universe->k);

    // Start writing the data of each cluster
    for (int i = 0; i < universe->k; i++) {
        // Prints the cluster symbol
        fprintf(file, CLUSTER_SYMBOL);
        fprintf(file, "\n");

        for (int j = 0; j < universe->nb_points; j++) {
            // Prints the point of the cluster
            if (universe->points_array[j]->cluster == universe->clusters_array[i]) {
                for (int d = 0; d < universe->dimensions; d++) {
                    fprintf(file, "%.2f", universe->points_array[j]->value[d]);
                    if (d < universe->dimensions - 1) {
                        fprintf(file, SEPARATOR);
                    } else {
                        fprintf(file, "\n");
                    }
                }
            }
        }
    }

    // Close and save the file
    fclose(file);
}

// Choose a random point by using the Fisher-Yates algorithm
point* choose_random_point_as_centroid(kmeans *universe, int* points_index_possible, int max_index) {
    int random_index = random_with_min_man_value(0, max_index);
    float* value = malloc(sizeof(float) * universe->dimensions);

    for (int d = 0; d < universe->dimensions; d++) {
        value[d] = universe->points_array[random_index]->value[d];
    }

    // Swap the selected point index with the last one,
    // so when max_index is decremented we can't choose again the same point
    swap(&points_index_possible[random_index], &points_index_possible[max_index]);

    point *centroid = point_create(value, universe->dimensions);
    return centroid;
}

void init_clusters(kmeans *universe) {
    // Create the index of each possible points that can be a centroid
    int* random_index_possible = malloc(sizeof(int) * universe->nb_points);
    for (int i = 0; i < universe->nb_points; i++) {
        random_index_possible[i] = i;
    }
    int max_index = random_index_possible[universe->nb_points - 1];

    // Choose a random centroid for each cluster
    for (int i = 0; i < universe->k; i++) {
        universe->clusters_array[i]->centroid = choose_random_point_as_centroid(universe, random_index_possible, max_index);
        max_index--;
    }

    free(random_index_possible);
}

float compute_distance(point* p1, point* p2) {
    if (p1->dimensions != p2->dimensions) {
        printf("The points don't have the same dimensions!\n");
        exit(EXIT_FAILURE);
    }
    float euclidean = compute_euclidean_distance(p1->value, p2->value, p1->dimensions);
    //float manhattan = compute_manhattan_distance(p1->value, p2->value, p1->dimensions);
    //float chebyshev = compute_chebyshev_distance(p1->value, p2->value, p1->dimensions);

    return euclidean;
}

bool compute_center_of_gravity(cluster* clstr, kmeans* universe) {
    bool new_position = false;
    int nb_points_in_cluster = 0;

    // Create an array to determine the center of gravity
    float* dimensions_average = malloc(sizeof(float) * universe->dimensions);
    for (int i = 0; i < universe->dimensions; i++) {
        dimensions_average[i] = 0;
    }

    // Parse each point in the cluster
    for (int i = 0; i < universe->nb_points; i++) {
        if (universe->points_array[i]->cluster == clstr) {
            // Compute their position
            for (int d = 0; d < universe->dimensions; d++) {
                dimensions_average[d] += universe->points_array[i]->value[d];
                nb_points_in_cluster += 1;
            }
        }
    }

    // Compute the center of gravity with the average position of each points in the cluster
    if (nb_points_in_cluster > 0) {
        for (int i = 0; i < universe->dimensions; i++) {
            dimensions_average[i] = dimensions_average[i] / nb_points_in_cluster;
            if (clstr->centroid->value[i] != dimensions_average[i]) {
                clstr->centroid->value[i] = dimensions_average[i];
                new_position = true;
            }
        }
    }

    free(dimensions_average);
    return new_position;
}

void assign_points_to_cluster(point* p, kmeans* universe) {
    float smallest_distance = 999;
    cluster* clst;
    for (int i = 0; i < universe->k; i++) {
         float distance = compute_distance(p, universe->clusters_array[i]->centroid);
         if (distance < smallest_distance) {
             smallest_distance = distance;
             clst = universe->clusters_array[i];
         }
    }
    p->cluster = clst;
}

void start_clustering(kmeans* universe) {
    bool clustering_in_progress = false;
    init_clusters(universe);

    do {
        clustering_in_progress = false;

        // Assign each points to their corresponding cluster
        for (int i = 0; i < universe->nb_points; i++) {
            assign_points_to_cluster(universe->points_array[i], universe);
        }

        // Compute the new center of gravity for each cluster
        printf("Clusters positions...\n");
        for (int i = 0; i < universe->k; i++) {
            if (compute_center_of_gravity(universe->clusters_array[i], universe)) {
                clustering_in_progress = true;
            }
            printf("Cluster %d position : ", i);
            for (int j = 0; j < universe->dimensions; j++) {
                printf("%0.2f, ", universe->clusters_array[i]->centroid->value[j]);
            }
            printf("\n");
        }
    } while (clustering_in_progress);
}

void destroy_point(point* p) {
    if (p != NULL) {
        if (p->label != NULL)
            free(p->label);

        if (p->value != NULL) {
            free(p->value);
        }

        free(p);
    }
}

void destroy_cluster(cluster* clstr) {
    if (clstr != NULL) {
        destroy_point(clstr->centroid);
        free(clstr);
    }
}

void destroy_universe(kmeans* kmeans) {
    if (kmeans != NULL) {
        if (kmeans->points_array != NULL) {
            for (int i = 0; i < kmeans->nb_points; i++) {
                destroy_point(kmeans->points_array[i]);
            }
        }
        if (kmeans->clusters_array != NULL) {
            for (int i = 0; i < kmeans->k; i++) {
                destroy_cluster(kmeans->clusters_array[i]);
            }
        }

        free(kmeans->points_array);
        free(kmeans->clusters_array);
        free(kmeans);
    }
}